#!/usr/bin/python3
#
# neq_terms.py - program for generating possible subshell configurations
# for nonequivalent electrons.
#
# Author: Adam Drzewiecki
# e-mail: A.Drzewiecki@if.uz.zgora.pl
# WWW: http://www.if.uz.zgora.pl/~adam
#
# Project finished (Python 2): 26.12.2011
# Python 2 to Python 3 conversion: 23.03.2019
# Last maintenance date: 25.03.2019
#

from sys import argv
from terms import *

#
# Working functions of the program
#

def get_l_text(number, test_args = False):
    """ Helper function for getting text representation of l quantum number

    If test_args is True, function returns second commandline argument if
    there are at least two after program name. In the other cases function
    returns string from console.
    """
    # Check second argument from the commandline on demand
    if test_args and len(argv) >= 2*number + 1:
        return argv[2*number - 1]
    else:
        return input("Input l quantum number or letter: ")

def parse_as_int(text):
    """ Helper function to convert text into l quantum number

    Function expects text form of integer literal. If cast to int will fail,
    function returns None.
    """
    try:
        return int(text)
    except ValueError:
        return None

def parse_as_symbol(text):
    """ Helper function to convert text into l quantum number

    Function expexts text in the form of spectroscopic L letter. Otherwise,
    function returns None.
    """
    try:
        return l_number(text)
    except ValueError:
        return None

def get_l(number):
    """ Function gets l number value, from comandline or keyboard
    """
    # Get textual form, including argument from commandline
    text = get_l_text(number, True)
    while True:
        # Try to interpret text input as an integer value
        l = parse_as_int(text)

        # If text is not integer value try to parse it as a spectroscopic symbol
        if l is None:
            l = parse_as_symbol(text)

        # If text was properly parsed and number is non-negative
        if l is not None and l >= 0:
            return l

        # Commandline argument considered only in the first iteration
        text = get_l_text(number)

def get_n_text(n_max, number, test_args = False):
    """ Helper function for getting text representation of electron number

    If test_args is True, function returns third commandline argument if
    there are at least two after program name. In the other cases function
    returns string from console.
    """
    # Check second argument from the commandline on demand
    if test_args and len(argv) >= 2*number + 1:
        return argv[2*number]
    else:
        return input("Input number of electrons (1-{0}): ".format(n_max))

def get_n(n_max, number):
    """ Function gets from keyboard electron number
    """
    # Get textual form, including argument from commandline
    text = get_n_text(n_max, number, True)
    while True:
        # try to convert text into integer value
        n = parse_as_int(text)

        # If conversion was succesful and number is OK, return it.
        if n is not None and 1 <= n <= n_max:
            return n

        # Get another text input only from console
        text = get_n_text(n_max, number)

def get_config():
    """ Function get_config() gathers subshell configurations from commandline or from keyboard input.
    """
    # result is a tuple of two element tuples, empty at start
    result = ()
    # index of analyzed subshell
    number = 1

    # Get one subshell configuration at once
    while True:
        # Get l number and number of electrons
        l = get_l(number)
        n = get_n(2*(2*l+1), number)

        # Append new subshell and update subshell index for the next iteration.
        result += (l, n),
        number += 1

        # Print some information on the screen - this is important for commandline errors correction.
        print("Subshell: {0}".format(sbs_conf_tostr((l, n))))

        # If there is not enough commandline arguments to parse as the subshell configurations
        if len(argv) == 1:
            # Repeat until user answer is 'y' or 'n'
            while True:
                # Ask user if there will be another subshell
                cont = input("Continue? [y/n]")
                if cont == 'y' or cont == 'n':
                    break

            # Stop the loop on demand
            if cont == 'n':
                break

        # ...or if there are possible subshell configurations in the commandline...
        elif len(argv) > 1:
            # Stop if there are not enough arguments left for another subshell.
            if len(argv) < 2*number + 1:
                break

    # Return gathered tuple of tuples
    return result

def show_microstates(sgrp_conf, microstates):
    """ Function shows list of microstates

    Function expect conf as tuple of a two element tuples (l, electron number)
    and microstates as a list of group of subshells configurations: tuple of
    subshells configurations:
    """
    if len(microstates) != 1:
        print("\nConfiguration {0} have {1} microstates:\n".format(sgrp_conf_tostr(sgrp_conf), len(microstates)))
    else:
        print("\nConfiguration {0} have one microstate:\n".format(sgrp_conf_tostr(sgrp_conf)))

    for state in microstates:
        print(sgrp_tostr(state) + \
            " Ml: {0}, Ms: {1}".format(sgrp_Ml(state), sgrp_Ms(state)))
    print()

def show_population_table(table):
    """ Function shows the given population table with header
    """
    print("Table of populations:\n")
    print_population_table(table)

def show_term_microstates(classified, mask):
    """ Function shows groups of microstates belonging to the given term

    Function expects classified as a 2D table of microstates lists, and mask
    as a table of the same geometry of integer type.
    """
    # Scan by rows
    j = 0
    while j < len(classified):
        # ...and columns
        i = 0
        while i < len(classified[j]):
            # If there are microstates marhed in mask table
            if mask[j][i] > 0:
                # Get list of microstates from table of classes
                s = classified[j][i]

                # Print header
                if len(s) > 1:
                    print("-) Ml: {0}, Ms: {1}, one of {2} following:".format(sgrp_Ml(s[0]), sgrp_Ms(s[0]), len(s)))
                elif len(s) == 1:
                    print("-) Ml: {0}, Ms: {1}".format(sgrp_Ml(s[0]), sgrp_Ms(s[0])))
                else:
                    raise ValueError("Inconsistence of classified and term mask!")

                # Print every microstate - subshell configuration
                for ms in classified[j][i]:
                    print_sgrp(ms)

                # Add some separator
                print()

            # Go to the next column
            i += 1

        # Go to the next row
        j += 1

def show_terms(terms, class_table):
    """ Function show information about terms detected in set of microstates
    """
    # Print header
    print("Decomposition to terms:\n")

    for term in terms:
        # Show term symbol
        print("Term: " + term_symbol(term[0]) + " (S: {0}, L: {1})\n".format(term[0][0], term[0][1]))

        # Show population table
        print("Population table:\n")
        print_population_table(term[1])

        # Show lists of microstates
        print("Microstates of term " + term_symbol(term[0]) + ":\n")
        show_term_microstates(class_table, term[1])

def show_summary(conf, terms):
    """ Function shows compact summary about found terms
    """
    print("Summary:\n")

    print("Configuration: {0}".format(sgrp_conf_tostr(conf)))
    print("Number of terms: {0}".format(len(terms)))
    print("List of terms:", end=" ")
    for term in terms:
        print(term_symbol(term[0]), end=" ")
    print()

#
# main()
#
def main():
    """ Main function of the program
    """
    print("Program \"neq_terms\" generates configurations of the given atomic subshell group\n")

    # Get subshell configurations
    config = get_config()

    # Prepare set of microstates of subshell
    microstates = sgrp_ustates(config)
    # Prepare table of classified microstates
    class_table = sgrp_classify_ustates(microstates)
    # Prepare population table for found classes
    populations = generate_population_table(class_table)
    # Split populations table into terms
    terms = decompose_population_table(populations)

    # Show microstates
    show_microstates(config, microstates)

    # Show population table
    show_population_table(populations)

    # Print out info about terms
    show_terms(terms, class_table)

    # Show summary
    show_summary(config, terms)

#
# Program launcher
#
if __name__ == "__main__":
    main()
