"""

Module "terms.py" consists of functions related to atomic terms analysis and classification.


Author: Adam Drzewiecki
e-mail: A.Drzewiecki@if.uz.zgora.pl
WWW: http://www.if.uz.zgora.pl/~adam

Project finished (Python 2): 30.10.2011
Python 2 to Python 3 conversion: 23.03.2019
Last maintenance date: 25.03.2019

"""

from fractions import Fraction

#
# Functions related to L quantum number naming convention
#

# L quantum number and its naming convention for l between 0 and 20
l_symbols = ("s",
             "p", "d", "f", "g", "h",
             "i", "k", "l", "m", "n",
             "o", "q", "r", "t", "u",
             "v", "w", "x", "y", "z")

def l_symbol(l_number):
    """Function l_symbol returns letter (s, p, ...) corresponding to l_number value.

    l_number must be integer between 0 and 20. Function returns small letters.

    >>> l_symbol(0)
    's'
    >>> l_symbol(1)
    'p'
    >>> l_symbol(2)
    'd'
    >>> l_symbol(3)
    'f'
    """
    l_number = int(l_number)
    
    if 0 <= l_number <= 20:
        return l_symbols[l_number]
    else:
        raise ValueError("l_number must be between 0 and 20, is: {0}".format(l_number))

def l_term_symbol(l_number):
    """ l_term_symbol return letter corresponding to L number for term.

    L number in term description returns capital letters.


    >>> l_term_symbol(0)
    'S'
    >>> l_term_symbol(1)
    'P'
    >>> l_term_symbol(2)
    'D'
    >>> l_term_symbol(3)
    'F'
    """
    l_number = int(l_number)
    
    return l_symbol(l_number).upper()

def l_number(l_symbol):
    """ Function l_number returns numeric L value for the given letter.

    l_symbol should be one letter string with spectroscopic L symbol.

    >>> l_number('S')
    0
    >>> l_number('s')
    0
    >>> l_number('P')
    1
    >>> l_number('p')
    1
    """
    # make function case insensitive
    l_lowered = l_symbol.lower()

    try:
        # try to find lowered symbol in symbols table
        return l_symbols.index(l_lowered)
    except ValueError:
        raise ValueError("l_symbol ({0}) not known".format(l_symbol))

#
# Functions related to electron state
#

def el_tostr(state):
    """ Function return string representation of electron state.

    Function expects one of following three values 1/2, 0, -1/2
    as an elecron state.

    >>> el_tostr(Fraction("1/2"))
    '+'
    >>> el_tostr(Fraction("-1/2"))
    '-'
    >>> el_tostr(Fraction(0))
    ' '
    """
    if state == Fraction("1/2"):
        return "+"
    elif state == Fraction("-1/2"):
        return "-"
    elif state == Fraction(0):
        return " "
    else:
        raise ValueError("Wrong electron state ({0}) detected".format(state))

def print_el(state):
    """ Function output electron state to screen.
    """
    print(el_tostr(state))

#
# Functions related to atomic terms
#

def term_symbol(term):
    """Function term_symbol returns string (multiplicity value)(L symbol).

    Function expects term in form of (S, L) tuple.
    """
    S = term[0]
    L = term[1]

    return "{0}{1}".format(int(2*S + 1), l_term_symbol(L))

#
# Function related to orbital configuration
#

def orb_el_number(orb):
    """Function returns number of electrons in the given orbital.

    Function expects orb is a two element tuple, each in (-1/2, 0, 1/2).
    Function depends on >>zero means lack of electron in position<< rule.
    """
    # Result aggregation starts from zero    
    result = 0

    # For every position on the orbital
    for position in orb:
        # Zero marks empty place in orbital configuration
        if position != 0:
            # If nonzero value, add one electron
            result += 1

    # Return aggregated result
    return result

def orb_Ml(orb, ml):
    """Function returns Ml for the given orbital.

    Function expects orb is a two element tuple, each in (-1/2, 0, 1/2),
    zero means "lack of electron". Ml is calculated according to given
    ml quantum number of orbital and number of electrones in it.
    """
    # Every electron on the orbital gives "ml" part to the Ml
    return ml * orb_el_number(orb)

def orb_Ms(orb):
    """Function returns Ms number of the given orbital.

    Function expects orb is a two element tuple, each in (-1/2, 0, 1/2),
    zero means "lack of electron". Ms is calculated as a sum of spins.
    "Zero means no electron" rule is crucial in this scheme.
    """
    # Every electron gives 1/2 or -1/2 to Ms
    return sum(orb)

def orb_tostr(orbital, separator = " "):
    """ Function returns string prepresenting orbital state.
    """
    # Starting with an empty string
    result = ""

    # For every electron in orbital    
    i = 0
    while i < len(orbital):

        # Add text representation of current electron to the result
        result += el_tostr(orbital[i])

        # If current electron is not the last one, add separator
        if i < len(orbital) - 1:
            result += separator

        # Go to next elecrtron
        i += 1

    # Return aggregated result
    return result

def print_orb(orbital):
    """ Function prints orbital state to the screen
    """
    print(orb_tostr(orbital))


#
# General functions related to microstates sets
#

def fun_extremes(ustates, fun):
    """ Function calculates minimum and maximum of fun(microstate) in their set

    As usual function expects at least one microstate in set. Result is returned
    as a tuple (minimum value, maximum value). fun must be able to work on one
    microstate in ustates set.
    """
    # We need at least one element
    if len(ustates) < 1:
        raise ValueError("ustates can not be an empty list")

    # Set value of fun(first element) as a starting value of extremes
    fun_min = fun_max = fun(ustates[0])

    # Let's start from the second element
    n = 1
    while n < len(ustates):
        # Get value of fun(current state)
        current = fun(ustates[n])

        # Test it!
        if current < fun_min:
            fun_min = current

        if current > fun_max:
            fun_max = current

        # Analyze next microstate on the list in the next iteration
        n += 1

    # Return aggregated extreme values an a tuple
    return (fun_min, fun_max)

#
# Function related to subshels
#

def sbs_Ml(sbs):
    """ Function returns Ml for the given atomic subshell.

    Function expects sbs is a tuple of 2*L+1 elements, being orbital, which
    is a two element tuple of (-1/2, 0, 1/2) (see orb_Ml function).
    """
    # Calculate L from number of orbitals in configuration
    L = (len(sbs) - 1) / 2

    # Result aggregation starts from zero
    result = 0

    # Number of current orbital in order
    n = 0

    # For every orbital in subshell
    for orb in sbs:
        # Add part of Ml number from current orbital
        #
        # Orbitals are treated as a sequence with ml = L, L-1, L-2,..., -L
        result += orb_Ml(orb, L - n)

        # Setup orbital index for the next iteration
        n += 1

    # Return aggregated result
    return result

def sbs_Ms(sbs):
    """ Function returns Ms for the given atomic subshell.

    Function expects sbs is a tuple of 2*L+1 elements, being orbital, which
    is a two element tuple of (-1/2, 0, 1/2) (see orb_Ms function).
    """
    # Result aggregation starts from zero 
    result = 0

    # For every orbital in subshell    
    for orb in sbs:
        # Add part of Ms number from current orbital
        result += orb_Ms(orb)

    # Return aggregated result
    return result

def sbs_generator(orb_nmb, el_nmb):
    """ Helper function sbs_generator generates list of subshell configurations - microstates.

    Function assumes orb_number as a number of orbitals and
    el_mnb as a number of electrons in them, and creates a list
    of all possible configurations.
    """
    #
    # Possible configurations of electron orbital occupation.
    # Order chosen according to Hund I ("bus seat rule")
    #
    configurations = (
        (Fraction("1/2"), Fraction(0)),
        (Fraction("-1/2"), Fraction(0)),
        (Fraction("1/2"), Fraction("-1/2")),
        (Fraction(0), Fraction(0))
        )

    # Empty list as a start of aggregating result
    result = []

    # One orbital base case
    if orb_nmb == 1:
        for cfg in configurations:
            # If configuration have proper number of electrons...
            if el_nmb == orb_el_number(cfg):
                # append it to the result as an one element tuple
                result.append((cfg,))

    # Many orbitals recursive case
    elif orb_nmb > 1:
        for cfg in configurations:
            # Create list of tails of the desired configuration,
            # tail have:
            # - one orbital less
            # - number of electrones - number in cfg
            #   (zero is OK for generating empty orbitals in sequence)
            #
            # Then for every obtained tail prepend current orbital - to
            # follow Hund II rule - and add to result list
            #
            if orb_el_number(cfg) <= el_nmb:    
                # Generate tails via recursive call
                tails = sbs_generator(orb_nmb - 1, el_nmb - orb_el_number(cfg))

                # for every tail
                for t in tails:
                    # Append (current orbital as an one element tuple + current tail) to the result 
                    result.append((cfg,) + t )

    # we shouldn't be here... orb < 1
    else:
        raise ValueError(
            "orb_nmb () must be integer >= 1".format(orb_nmb))

    # Return aggregated list of microstates
    return result

def sbs_ustates(sbs_conf):
    """ Function sbs_conf generates list of subshell configurations - microstates

    Function expects sbs_conf as a two element tuple: (l, number of electrons)
    """
    return sbs_generator(2*sbs_conf[0] + 1, sbs_conf[1])

def sbs_classify_ustates(sbs_ustates):
    """ Function splits microstates into groups of the same Ms and Ml.

    Result of a function is list[same Ml][same Ms][sbs_microstates].
    """
    # Find minimum and maximum of Ms in the given set of microstates
    (min_ms, max_ms) = fun_extremes(sbs_ustates, sbs_Ms)
    # Calculate number of columns for S coordinate
    columns = int(max_ms - min_ms) + 1

    # Find minimum and maximum od Ml in the given set of microstates
    (min_ml, max_ml) = fun_extremes(sbs_ustates, sbs_Ml)
    # Calculate number of rows for L coordinate
    rows = int(max_ml - min_ml) + 1

    # Prepare result[rows][columns][] for aggregating subsets of microstates
    result = [[[] for c in range(int(columns))] for r in range(int(rows))]

    # For every subshell microstate in a list
    for sbs_ustate in sbs_ustates:
        # Calculate Ml for current microstate
        current_ml = sbs_Ml(sbs_ustate)
        # Calculate row index according to Ml
        index_ml = int(max_ml - current_ml)

        # Calculate Ms for current microstate
        current_ms = sbs_Ms(sbs_ustate)
        # Calculate column index according to Ms
        index_ms = int(max_ms - current_ms)

        # Append microstate at result[index_ml][index_ms]
        result[index_ml][index_ms].append(sbs_ustate)

    # Return aggregated list of results
    return result

def sbs_tostr(sbs, prefix="|", sufix="|", separator="|"):
    """ Function returns string representation of subshell state.
    """
    # Start with an empty string
    result = ""

    # Get the number of orbitals in subshell
    orb_nmb = len(sbs)

    # Add opening sign
    if orb_nmb > 0:
        result += prefix

    # For every orbital in subshell
    i = 0
    while i < orb_nmb:
        # Add description of i-th orbital in subshell
        result += orb_tostr(sbs[i])

        # If it was not the last orbital add separator
        if i < orb_nmb - 1:
            result += separator

        # Get next orbital
        i += 1

    # Add closing sign
    if orb_nmb > 0:
        result += sufix

    # Return aggregated result
    return result

def print_sbs(state):
    """ Function prints subshell state to the screen
    """
    print(sbs_tostr(state))

def sbs_conf_tostr(conf):
    """ Function returns subshell configuration string: (l symbol)(electrons)

    Function expects conf as a two element tuple: (l, number of electrons)
    """
    try:
        return "{0}{1}".format(l_symbol(conf[0]), conf[1])
    except IndexError:
        raise ValueError("conf must be two element tuple")

def print_sbs_conf(conf):
    """ Function prints subshell configuration to the screen
    """
    print(sbs_conf_tostr(conf))

#
# Functions related to subshell groups
#

def sgrp_Ml(sgrp):
    """ Function returns Ml for the given group of atomic subshell.

    Function expects sgrp is a tuple of 2*L+1 elements tuples of orbital type,
    which is a two element tuple of (-1/2, 0, 1/2) (see orb_Ml function).
    """
    # Proper Ml value should be of integer type
    result = 0

    # Ml is counted as a sum of Ml numbers for each subshaell
    for sbs in sgrp:
        result += sbs_Ml(sbs)

    # Return aggregated result
    return result

def sgrp_Ms(sgrp):
    """ Function returns Ms for the given group of atomic subshell.

    Function expects sgrp is a tuple of 2*L+1 elements tuples of orbital type,
    which is a two element tuple of (-1/2, 0, 1/2) (see orb_Ms function).
    """
    # Proper Ms value should be of Fraction type (multiple of 1/2)
    result = Fraction(0)

    # Ms is counted as a sum of Ms numbers for each subshaell
    for sbs in sgrp:
        result += sbs_Ms(sbs)

    # Return aggregated result
    return result

def sgrp_ustates(sgrp_conf):
    """ Function generates list of subshells group configurations - microstates.

    Function assumes sgrp_conf as a tuple of two element tuples: orbitals number
    in a subshell and and number of electrons on it, and creates a list of all
    possible configurations.
    """
    # Start with an empty list
    result = []

    # How many subshells?
    subshells = len(sgrp_conf)

    # If demanded number of subshells is at least one, prepare first set
    if subshells >= 1:
        first_set = sbs_ustates(sgrp_conf[0])

    # One subshell case
    if subshells == 1:
        # Prepare result list in the form of list of one element tuples
        for first in first_set:
            # Promote one subshell configuration into one element group
            result.append( (first,) )

    # Many subshells recursive case
    elif subshells > 1:
        # Prepare configurations of all subshells except first
        second_set = sgrp_ustates(sgrp_conf[1:])

        # Make cartesian product: take every element from the first set
        for first in first_set:
            # .. and for every element from the second set
            for second in second_set:
                # Add to the result joined tuples as a new element
                result.append( (first,) + second )

    # Return aggregated result
    return result

def sgrp_classify_ustates(sgrp_ustates):
    """ Function splits microstates into groups of the same Ms and Ml.

    Result of a function is list[same Ml][same Ms][sgrp_microstates].
    """
    # Find minimum and maximum of Ms in the given set of microstates
    (min_ms, max_ms) = fun_extremes(sgrp_ustates, sgrp_Ms)
    # Calculate number of columns for S coordinate
    columns = int(max_ms - min_ms) + 1

    # Find minimum and maximum od Ml in the given set of microstates
    (min_ml, max_ml) = fun_extremes(sgrp_ustates, sgrp_Ml)
    # Calculate number of rows for L coordinate
    rows = int(max_ml - min_ml) + 1

    # Prepare result[rows][columns][] for aggregating subsets of microstates
    result = [[[] for c in range(int(columns))] for r in range(int(rows))]

    # For every subshell microstate in a list
    for sgrp_ustate in sgrp_ustates:
        # Calculate Ml for current microstate
        current_ml = sgrp_Ml(sgrp_ustate)
        # Calculate row index according to Ml
        index_ml = int(max_ml - current_ml)

        # Calculate Ms for current microstate
        current_ms = sgrp_Ms(sgrp_ustate)
        # Calculate column index according to Ms
        index_ms = int(max_ms - current_ms)

        # Append microstate at result[index_ml][index_ms]
        result[index_ml][index_ms].append(sgrp_ustate)

    # Return aggregated list of results
    return result

def sgrp_tostr(sgrp, separator=" "):
    """ Function returns string representation of group of subshells state.
    """
    # Start with an empty string
    result = ""

    # Get the number of orbitals in subshell
    sbs_nmb = len(sgrp)

    # For every subshell in a group
    i = 0
    while i < sbs_nmb:
        # Add description of i-th subshell in the group
        result += sbs_tostr(sgrp[i], prefix="<", sufix=">")

        # If it was not the last subshell add separator
        if i < sbs_nmb - 1:
            result += separator

        # Get next subshell
        i += 1

    # Return aggregated result
    return result

def print_sgrp(sgrp):
    """ Function prints state of group of subshells to the screen
    """
    print(sgrp_tostr(sgrp))

def sgrp_conf_tostr(sgrp_conf):
    """ Function returns subshell group configuration string:
    (l symbol)(electrons) for every subshell

    Function expects conf as tuple of a two element tuples:
    (l, number of electrons)
    """
    # Start with an empty string
    result = ""

    try:
        # For every subshell configuration add two element part to the result
        for conf in sgrp_conf:
            result += sbs_conf_tostr(conf)
    except IndexError:
        raise ValueError("conf must be two element tuple")

    # Return aggregated result
    return result

def print_sgrp_conf(sgrp_conf):
    """ Function prints subshell configuration to the screen
    """
    print(sgrp_conf_tostr(sgrp_conf))

#
# Functions related to population tables
#

def L_from_geometry(rows):
    """ Function calculates value of L quantum number from number of rows.

    This is a helper function related to population table functions.
    """
    # Expected int value as a result
    return (rows - 1) / 2

def S_from_geometry(cols):
    """ Function calculates value of S quantum number from number of columns.

    This is a helper function related to population table functions.
    """
    # Expected integer multiplicity of 0.5 as a result
    return float(cols - 1) / 2.0

def generate_population_table(classified_ustates):
    """ Function counts states in splitted list of microstates.

    Function rely on assumption of
    classified_microstate[same Ml][same Ms][any microstates]
    structure of argument. Function returns list[same Ml][same Ms][int]
    """
    # Prepare list of list of numbers of states with the same Ml and Ms
    result = [[len(col) for col in row] for row in classified_ustates]

    # Return aggregated result
    return result

def subtract_population_table(left, right):
    """ Function substract right population table from left population table

    Function expects both table of the same dimensions.
    """
    try:
        # Prepare result by scanning row by row and column by column
        result = [ [left[r][c] - right[r][c] for c in range(len(left[r]))]
            for r in range(len(left)) ]
    except IndexError:
        raise ValueError("left and right must have the same dimensions")

    # Return aggregated result
    return result

def term_population_table(rows, cols, L, S):
    """ Function prepares population table of given geometry for specified term.

    Rows and columns must be positive integers. L and S must be non-negative
    integers, considered as quantum numbers of term, described by generated
    population table.
    """
    # Function can not wor properly with wrong values of rows and cols
    if rows < 1 or cols < 1:
        raise ValueError(
        "rows ({}) and cols ({}) must be greater or equal 1".format(rows, cols))

    # Calculate maximum values of L and S from target table geometry
    Lmax = L_from_geometry(rows)
    Smax = S_from_geometry(cols)

    # Prepare zero filled table of int
    result = [[0 for c in range(cols)] for r in range(rows)]

    # Proper values of L and S are crucial
    if L < 0 or S < 0:
        raise ValueError("L ({0}) and S ({1}) must be at least zero".format(L, S))

    # For rows from Lmax - L to Lmax + L
    j = int(Lmax - L)
    while(j <= Lmax + L):

        # For columns from Smax - S to Smax + S
        i = int(Smax - S)
        while(i <= Smax + S):

            # Set value to 1
            result[j][i] = 1

            # Next column
            i += 1

        # Next row
        j += 1

    # Return prepared table
    return result

def decompose_population_table(table):
    """ Function factorises population table into terms population tables

    Function expects proper population table, generated from set of categorized
    microstates by generate_population_table as an argument. Result is a list
    of two element tuples. First element is a (S, L) tuple with quantum numbers
    of found term. Second element is a population table of this term.
    """
    # Starting from an empty list
    result = []

    # Find geometry of table 
    rows = len(table)
    cols = len(table[0])

    # Rows means populations of microstates with ML from Lmax to -Lmax
    Lmax = L_from_geometry(rows)
    # Columns means populations of microstates with Ms from Smax to -Smax
    Smax = S_from_geometry(cols)

    # Number of rows and columns to consider (upper left quarter is scanned):
    # half of all + 1 (middle one) for odd number
    row_max = rows // 2 + rows % 2
    col_max = cols // 2 + cols % 2

    # Scan in order: by columns then by rows to autosort terms by Hund I and II
    c = 0
    while(c < col_max):

        r = 0
        while(r < row_max):

            # If there is at least one microstate
            if table[r][c] > 0:

                # Calculate current value of L and S for row and column
                Lc = Lmax - r
                Sc = Smax - c

                # Prepare term population table
                right = term_population_table(rows, cols, Lc, Sc)

                # Substract term table until there are proper microstates;
                # sometimes there are few of the same terms in table
                while table[r][c] > 0:

                    # Subtract term population table (right) from table
                    table = subtract_population_table(table, right)

                    # Append newly found term to result
                    result.append( ((Fraction.from_float(Sc), Lc), right) )

            # Go to next row
            r += 1

        # Go to next column
        c += 1

    # Return aggregated result
    return result

def population_table_tostr(table):
    """ Function return string representation of population table
    """
    # Get L and S from geometry of population table
    L = L_from_geometry(len(table))
    S = S_from_geometry(len(table[0]))

    # Starting from empty string
    result = ""

    # Prepare header of the table

    # Insert filler for L symbols
    result += "  | "
    # For every column of population table
    for i in range(len(table[0])):
        # Insert multiplicity number for S in current column
        result += "{0:3} ".format(int(2*abs(S - i) + 1))
    # End up header with newline character
    result += "\n"
    # Add separator line between header and contents of the population table
    result += (3 + len(table[0])*4)* "-" + "\n"

    # Prepare contents of the table

    # n is scanned between L and -L
    n = L
    for row in table:
        # Prepare description of L quantum number for the row
        result += l_term_symbol(abs(n))+ " | "

        # Prepare row of the table
        for col in row:
            # If there are microstates
            if col != 0:
                # Print their population
                result += "{0:3} ".format(col)
            else:
                # Skip printing zeros in table to improve readability
                result += "{0:3} ".format("")

        # End up line with newline character
        result += "\n"
        # Update L quantum number for next row
        n -= 1

    # Return aggregated result
    return result

def print_population_table(table):
    """ Function prints population table to the screen
    """
    print(population_table_tostr(table))

#
# Documentation test suite feature
#
if __name__ == "__main__":
    import doctest
    doctest.testmod()
